//---------------------------------------------------------------------------
//
//  Grammar loading and parsing support routines
//  Dmitry Brant, Spr 2004
//
//-----------

#include <vcl.h>
#pragma hdrstop

#include <stdio.h>
#include <stdlib.h>
#include <stdexcept.h>
#include "grammar.h"

using std::runtime_error;


//Damn it, Jim, I tried to save him!
void Grammar::LoadFromFile(String fileName){
    //destroy previous rules
    for(int i=0; i<ruleGroups->Count; i++) delete ((RewriteRuleGroup*)ruleGroups->Items[i]);
    ruleGroups->Clear();
    //load new file
    FILE* f = fopen(fileName.c_str(), "rb");
    if(!f) throw runtime_error("Failed to open grammar file");
    fseek(f, 0, SEEK_END);
    int fileLen = ftell(f);
    fseek(f, 0, SEEK_SET);
    char* buffer = new char[fileLen];
    try{
        fread(buffer, 1, fileLen, f);
        fclose(f);
        int bufPtr=-1, linePtr, lineLength;
        char line[1024];
        while(bufPtr < (fileLen-1)){
            linePtr=0;
            while(1){
                bufPtr++;
                if((bufPtr >= fileLen) || (buffer[bufPtr] == '\n') || (buffer[bufPtr] == '\r')) break;
                line[linePtr++] = buffer[bufPtr];
            }
            line[linePtr] = 0;
            lineLength = linePtr;

            //read the nonterminal
            String nonTerminal;
            linePtr = 0;
            while((line[linePtr] != '<') && (linePtr < lineLength)){
                if(line[linePtr] == '#') break; //it's a comment
                if(line[linePtr] > ' ') throw runtime_error("Grammar is not context-free");
                linePtr++;
            }
            if(line[linePtr] != '<') continue;  //allow for comments
            linePtr++;
            while((line[linePtr] != '>') && (linePtr < lineLength)){
                if(line[linePtr] > ' ') nonTerminal += String(line[linePtr]);
                linePtr++;
            }
            if(line[linePtr] != '>') continue;
            linePtr++;

            //look for the arrow symbol
            while((line[linePtr] != '-') && (linePtr < lineLength)){
                if(line[linePtr] > ' ') throw runtime_error("Grammar syntax error");
                linePtr++;
            }
            if(line[linePtr] != '-') continue;
            linePtr++;
            while((line[linePtr] != '>') && (linePtr < lineLength)){
                if(line[linePtr] > ' ') throw runtime_error("Grammar syntax error");
                linePtr++;
            }
            if(line[linePtr] != '>') continue;
            linePtr++;
            //got the arrow symbol; now read the replacement

            //but first, decide which group this production belongs to
            RewriteRuleGroup* theGroup=NULL;
            for(int i=0; i<ruleGroups->Count; i++){
                RewriteRuleGroup* group = (RewriteRuleGroup*)ruleGroups->Items[i];
                if(group->nonTerminal == nonTerminal){
                    theGroup = group; break;
                }
            }
            if(!theGroup){
                //ok, create a new group
                theGroup = new RewriteRuleGroup();
                ruleGroups->Add(theGroup);
                theGroup->nonTerminal = nonTerminal;
            }

            String rewrite;
            while(1){
                //go to beginning of rewrite
                while((line[linePtr] <= ' ') && (linePtr < lineLength)){
                    linePtr++;
                }
                if(linePtr >= lineLength) break;

                //Add a rule to the list
                RewriteRule* theRule = new RewriteRule();
                theGroup->rules->Add(theRule);

                //read the rewrite
                while((line[linePtr] > ' ') && (line[linePtr] != '|') && (linePtr < lineLength)){

                    rewrite = "";
                    
                    RewriteUnit* theUnit = new RewriteUnit();
                    theRule->units->Add(theUnit);

                    if(line[linePtr] == '<'){
                        //it's a nonterminal
                        theUnit->terminal = false;
                        linePtr++;
                        while((line[linePtr] != '>') && (linePtr < lineLength)){
                            if(line[linePtr] > ' '){
                                rewrite += String(line[linePtr]);
                            }
                            linePtr++;
                        }
                    }
                    else{
                        //it's a terminal
                        theUnit->terminal = true;
                        rewrite += String(line[linePtr]);
                    }
                    if((rewrite == "") || (rewrite == "lambda") || (rewrite == "epsilon")){ rewrite = ""; theUnit->terminal = true; }
                    theUnit->str = rewrite;
                    if((rewrite == "#gen")) theUnit->codeGen = true;
                    linePtr++;
                    while(((line[linePtr] == ' ') || (line[linePtr] == '\t')) && (linePtr < lineLength)) linePtr++;
                }
                if(line[linePtr] == '|'){ linePtr++; continue; }
                break;
            }
        };
    }__finally{
        //done!
        delete [] buffer;
    }
}
//---------------------------------------------------------------


void Grammar::SaveToFile(String fileName){
    FILE* f = fopen(fileName.c_str(), "w");
    if(!f) throw runtime_error("Failed to save the specified file");
    for(int i=0; i<ruleGroups->Count; i++){
        String currentStr;
        RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];
        currentStr += "<" + currentGroup->nonTerminal + "> -> ";
        for(int j=0; j<currentGroup->rules->Count; j++){
            if(j>0) currentStr += " | ";
            RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
            for(int k=0; k<currentRule->units->Count; k++){
                RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[k];
                if(unit->terminal){
                    if(unit->str == ""){
                        currentStr += "<epsilon>";
                    }else{
                        currentStr += unit->str;
                    }
                }else{
                    currentStr += "<" + unit->str + ">";
                }
            }
        }
        currentStr += "\r\n";
        fprintf(f, currentStr.c_str());
    }
    fclose(f);
}
//---------------------------------------------------------------


void Grammar::EliminateLeftRecursion(){

    //Algorithm:
    //Arrange all nonterminals in some arbitrary order: <A_1>, <A_2>, ..., <A_n>
    //For each nonterminal <A_i> {
    //  For each nonterminal <A_i> such that 1 <= j < i and there exists a production rule
    //  of the form <A_i> -> <A_j> R0, where the <A_j> productions are <A_j> -> R1 | R2 | ... | Rn {
    //    Replace the production rule <A_i> -> <A_j> R0 with <A_i> -> R1 R0 | R2 R0 | ... | Rn | R0
    //  }
    //  Eliminate any immediate left recursion among the <A_i> productions
    //}

    //Our list of nonterminals is simply the list of groups. Easy enough
    int numGroups = ruleGroups->Count;
    for(int w=0; w<numGroups; w++){
        RewriteRuleGroup* activeGroup = (RewriteRuleGroup*)ruleGroups->Items[w];

        for(int i=1; i<numGroups; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];

            //look for production rules where the active group references a nonterminal
            //from a lower-index group.

            for(int j=0; j<i; j++){
                RewriteRuleGroup* otherGroup = (RewriteRuleGroup*)ruleGroups->Items[j];


                for(int k=0; k<currentGroup->rules->Count; k++){
                    RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[k];
                    RewriteUnit* firstUnit = (RewriteUnit*)currentRule->units->Items[0];

                    if((!firstUnit->terminal) && (firstUnit->str == otherGroup->nonTerminal)){
                        //got it

                        for(int m=0; m<otherGroup->rules->Count; m++){
                            RewriteRule* otherRule = (RewriteRule*)otherGroup->rules->Items[m];
                            RewriteRule* newRule = new RewriteRule();
                            currentGroup->rules->Add(newRule);

                            for(int n=0; n<otherRule->units->Count; n++){
                                RewriteUnit* unit = (RewriteUnit*)otherRule->units->Items[n];
                                RewriteUnit* newUnit = new RewriteUnit();
                                newUnit->terminal = unit->terminal;
                                newUnit->codeGen = unit->codeGen;
                                newUnit->str = unit->str;
                                newRule->units->Add(newUnit);
                            }

                            for(int n=1; n<currentRule->units->Count; n++){
                                RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[n];
                                RewriteUnit* newUnit = new RewriteUnit();
                                newUnit->terminal = unit->terminal;
                                newUnit->codeGen = unit->codeGen;
                                newUnit->str = unit->str;
                                newRule->units->Add(newUnit);
                            }
                        }

                        delete currentRule;
                        currentGroup->rules->Delete(k);
                        k--;

                    }
                }
            }
        }

        //eliminate direct left-recursion in the current group
        bool directRecursionExists=false;
        for(int k=0; k<activeGroup->rules->Count; k++){
            RewriteRule* currentRule = (RewriteRule*)activeGroup->rules->Items[k];
            RewriteUnit* firstUnit = (RewriteUnit*)currentRule->units->Items[0];
            if((!firstUnit->terminal) && (firstUnit->str == activeGroup->nonTerminal)){
                //got it
                directRecursionExists = true;
            }
        }

        if(directRecursionExists){
            //create a new group!
            RewriteRuleGroup* newGroup = new RewriteRuleGroup();
            ruleGroups->Add(newGroup);
            newGroup->nonTerminal = (activeGroup->nonTerminal + "'");

            for(int k=0; k<activeGroup->rules->Count; k++){
                RewriteRule* currentRule = (RewriteRule*)activeGroup->rules->Items[k];
                RewriteUnit* firstUnit = (RewriteUnit*)currentRule->units->Items[0];
                if((!firstUnit->terminal) && (firstUnit->str == activeGroup->nonTerminal)){
                    //got it
                    newGroup->rules->Add(currentRule);
                    delete firstUnit;
                    currentRule->units->Delete(0);
                    RewriteUnit* newUnit = new RewriteUnit();
                    newUnit->str = newGroup->nonTerminal;
                    newUnit->terminal = false;
                    currentRule->units->Add(newUnit);
                    activeGroup->rules->Delete(k);
                    k--;
                    if(!activeGroup->rules->Count){
                        RewriteRule* newRule = new RewriteRule();
                        activeGroup->rules->Add(newRule);
                        RewriteUnit* newUnit = new RewriteUnit();
                        newRule->units->Add(newUnit);
                        newUnit->terminal = false;
                        newUnit->str = newGroup->nonTerminal;
                        k++;
                    }
                }else{
                    RewriteUnit* newUnit = new RewriteUnit();
                    newUnit->str = newGroup->nonTerminal;
                    newUnit->terminal = false;
                    currentRule->units->Add(newUnit);
                }
            }
            //must add an epsilon production to the new group
            RewriteRule* newRule = new RewriteRule();
            newGroup->rules->Add(newRule);
            RewriteUnit* newUnit = new RewriteUnit();
            newUnit->terminal = true;
            newUnit->str = "";
            newRule->units->Add(newUnit);
        }

        FixDegenerates();
    }
}
//---------------------------------------------------------------


void Grammar::EliminateEpsilonProductions(){

    //Algorithm:
    //While exists a production rule of the form <A_i> -> epsilon {
    //  if epsilon is the only item in the production group <A_i>, then the nonterminal
    //  qualifies for full elimination.
    //  Eliminate the production <A-i> -> epsilon.
    //  For each production rule of the form <A_k> -> R_1 <A_i> R_2 {
    //    For each occurrence of <A_i> in r.h.s of <A_k> {
    //      If full elimination is in effect, reduce the production rule to <A_k> -> R_1 R_2
    //      Else, add the production rule <A_k> -> R_1 R_2
    //    }
    //  }
    //}                 

    while(1){

        String offendingNonTerminal;
        bool exists=false, fullElimination;
        for(int i=0; i<ruleGroups->Count; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];
            for(int j=0; j<currentGroup->rules->Count; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
                for(int k=0; k<currentRule->units->Count; k++){
                    RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[k];
                    if(unit->str == ""){
                        //got one!
                        exists = true;
                        offendingNonTerminal = currentGroup->nonTerminal;
                        fullElimination = (currentGroup->rules->Count == 1);
                        //it's safe to delete the rewrite rule right here
                        if(fullElimination){
                            delete currentGroup;
                            ruleGroups->Delete(i);
                            i--;
                        }else{
                            delete currentRule;
                            currentGroup->rules->Delete(j);
                            j--;
                        }
                        break;
                    }
                }
                if(exists) break;
            }
            if(exists) break;
        }
        if(!exists) break;    //done!!

        for(int i=0; i<ruleGroups->Count; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];
            for(int j=0; j<currentGroup->rules->Count; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
                for(int k=0; k<currentRule->units->Count; k++){
                    RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[k];
                    if((!unit->terminal) && (unit->str == offendingNonTerminal)){
                        if(fullElimination){
                            //only collapse the current rule
                            if(currentRule->units->Count > 1){
                                delete unit;
                                currentRule->units->Delete(k);
                                k--;
                            }else{
                                unit->terminal = true;
                                unit->str = "";
                            }
                        }else{
                            //build a new rule
                            RewriteRule* newRule = new RewriteRule();
                            currentGroup->rules->Add(newRule);
                            if(currentRule->units->Count > 1){
                                for(int m=0; m<currentRule->units->Count; m++){
                                    if(m != k){
                                        RewriteUnit* myUnit = (RewriteUnit*)currentRule->units->Items[m];
                                        RewriteUnit* newUnit = new RewriteUnit();
                                        newRule->units->Add(newUnit);
                                        newUnit->str = myUnit->str;
                                        newUnit->terminal = myUnit->terminal;
                                        newUnit->codeGen = myUnit->codeGen;
                                    }
                                }
                            }else{
                                RewriteUnit* newUnit = new RewriteUnit();
                                newRule->units->Add(newUnit);
                                newUnit->str = "";
                                newUnit->terminal = true;
                            }
                        }
                    }
                }
            }
        }

        FixDegenerates();
    }
}
//---------------------------------------------------------------


void Grammar::FixDegenerates(){

    //Algorithm:
    //Look for the following anomalies, and repair:
    //- An empty rewrite rule group
    //- An empty rewrite rule
    //- Duplicate rewrite rules in one group
    //- A non-standalone epsilon unit in a rewrite rule

    bool checkAgain=true;
    while(checkAgain){

        checkAgain = false;
        for(int i=0; i<ruleGroups->Count; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];

            //empty group?
            if(!currentGroup->rules->Count){
                checkAgain = true;
                delete currentGroup;
                ruleGroups->Delete(i);
                i--;
                continue;
            }

            for(int j=0; j<currentGroup->rules->Count; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];

                //empty rule?
                if(!currentRule->units->Count){
                    checkAgain = true;
                    delete currentRule;
                    currentGroup->rules->Delete(j);
                    j--;
                    continue;
                }

                for(int k=0; k<currentRule->units->Count; k++){
                    RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[k];
                    if((unit->str == "") && (currentRule->units->Count > 1)){
                        //a stray epsilon!
                        checkAgain = true;
                        delete unit;
                        currentRule->units->Delete(k);
                        k--;
                    }
                }
            }

            //check for dupes
            for(int j=0; j<currentGroup->rules->Count-1; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
                for(int k=j+1; k<currentGroup->rules->Count; k++){
                    RewriteRule* otherRule = (RewriteRule*)currentGroup->rules->Items[k];
                    if(currentRule->equalTo(otherRule)){
                        delete otherRule;
                        currentGroup->rules->Delete(k);
                        k--;
                    }
                }
            }

        }

    }//while
}
//---------------------------------------------------------------


void Grammar::EliminateUnitProductions(){

    //Algorithm:
    //While exist production rules of the form <A_i> -> <A_j> {
    //  For each production rule of the form <A_i> -> <A_j> {
    //    Transfer the productions of the nonterminal <A_j> to the nonterminal <A_i>
    //    Remove the nonterminal <A_j>
    //    Replace all occurrences of <A_j> in all other productions with <A_i>
    //  }
    //}
    //This algorithm also implicitly removes cycles.
    //Prerequisite for this algorithm: remove epsilon-productions.

    while(1){

        RewriteRuleGroup* activeGroup;
        String rightNonTerminal;
        bool exists=false;
        for(int i=0; i<ruleGroups->Count; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];
            for(int j=0; j<currentGroup->rules->Count; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
                if(currentRule->units->Count == 1){
                    RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[0];
                    if(!unit->terminal){
                        //got one!
                        exists = true;
                        activeGroup = currentGroup;
                        rightNonTerminal = unit->str;
                        if(activeGroup->nonTerminal == rightNonTerminal){
                            //damn it, it's a direct cycle!
                            printf(("Cycle detected and destroyed. (from " + activeGroup->nonTerminal + " to " + activeGroup->nonTerminal + ")\n").c_str());
                            delete currentRule;
                            currentGroup->rules->Delete(j);
                            j--;
                            exists = false;
                        }
                    }
                }
                if(exists) break;
            }
            if(exists) break;
        }
        if(!exists) break;    //done!!

        for(int i=0; i<ruleGroups->Count; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];
            if(currentGroup->nonTerminal != rightNonTerminal) continue;
            for(int j=0; j<currentGroup->rules->Count; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
                activeGroup->rules->Add(currentRule);
            }
            currentGroup->rules->Clear();
            delete currentGroup;
            ruleGroups->Delete(i);
            break;
        }

        for(int i=0; i<ruleGroups->Count; i++){
            RewriteRuleGroup* currentGroup = (RewriteRuleGroup*)ruleGroups->Items[i];
            for(int j=0; j<currentGroup->rules->Count; j++){
                RewriteRule* currentRule = (RewriteRule*)currentGroup->rules->Items[j];
                for(int k=0; k<currentRule->units->Count; k++){
                    RewriteUnit* unit = (RewriteUnit*)currentRule->units->Items[k];
                    if((!unit->terminal) && (unit->str == rightNonTerminal)){
                        unit->str = activeGroup->nonTerminal;
                    }
                }
            }
        }

        FixDegenerates();
    }
}
//----------------------------------------------------------------------


